/*
autor:  Marc-Antoine Drouin
  email: drouim@iro.umontreal.ca

  description: stuff to open an image
  
  Copyright (c) 2005 Marc-Antoine Drouin
  Copyright (c) 1999-2005 Sbastien Roy

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/





/***** Define this for closest point instead of bi-linear interpolation *****/
/*#define CLOSEST_POINT*/

/*** Support either pbm or sgi format ***/

#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "img.h"


/* Define to inverse all Y in image */
#define INVERSE_Y

/** Define to have verbose info **/
#define VERBOSE


int CleanImage(imginfo *I, unsigned char col) 
{
  int y;
  
  for(y=0;y< IMYSIZE*IMXSIZE*IMZSIZE;y++)
    {
      I->Data[y ]= col;
    }
  return 0;
}




/********** PBM FORMAT SECTION ***********/

/** return 0 if ok, -1 if error **/
/** return the data in PData, and the size in P[XYZ]Size **/
/** ZSize=1 -> monochrome, ZSize=3 or 4 -> RGB or ARGB **/
/** If RGB : rgbrgbrgbrgb... **/
int LoadImage(char *Name,imginfo *I)
{
FILE *img;
unsigned char *Data;
/* Data : [XSize*YSize*ZSize)], Acces=[(Y*XSize+X)(*Zsz)+(z)] */
int P,x,y,z,v;
int XSize,YSize,ZSize,Ascii;
char TBuf[100];

 Data= I->Data;
 

	/* default values */



        img=fopen(Name,"r");
        if( img==NULL ) {
                fprintf(stderr,"LoadImage: Unable to open '%s'\n",Name);return(-1);
        }

	/** Read header **/
	if( fgets(TBuf,100,img)!=TBuf ) {
                fprintf(stderr,"LoadImage: Unable to read header of '%s'\n",Name);
		return(-1);
	}
	if( strncmp(TBuf,"P5",2)==0 ) { Ascii=0;ZSize=1; }	/* pgm raw */
	else if( strncmp(TBuf,"P6",2)==0 ) { Ascii=0;ZSize=3; }	/* ppm raw */
	else if( strncmp(TBuf,"P2",2)==0 ) { Ascii=1;ZSize=1; }	/* pgm ascii */
	else if( strncmp(TBuf,"P3",2)==0 ) { Ascii=1;ZSize=3; }	/* ppm ascii */
	else {
		fprintf(stderr,"LoadImage: Unsupported PNM format for '%s'\n",Name);
		return(-1);
	}
	/* Skip comments */
	do {
		if( fgets(TBuf,100,img)!=TBuf ) return(-1);
	} while( TBuf[0]=='#' );
	/* xres/yres */
	sscanf(TBuf,"%d %d",&XSize,&YSize);
	/* Number of gray , usually 255, Just skip it */
	if( fgets(TBuf,100,img)!=TBuf ) return(-1);

        /* print a little info about the image */
#ifdef VERBOSE
        printf("Image x and y size in pixels: %d %d\n"
                ,XSize,YSize);
        printf("Image zsize in channels: %d\n",ZSize);
        printf("Image pixel min and max: %d %d\n",0,255);
#endif

        /* Allocate data space , as char data... for now */

	if( YSize != IMYSIZE || XSize !=IMXSIZE)
	  {
	    
	    printf("image size not coresponding to const.h (%d %d %d) \n",XSize,YSize,ZSize);
	    return -1;
	  }
	



	if( Ascii==0 ) {
		for(y=0;y<IMYSIZE;y++) {
#ifdef INVERSE_Y
			P=(IMYSIZE-1-y)*IMXSIZE*ZSize;
#else
			P=y*IMXSIZE*ZSize;
#endif
			fread(Data+P,1,IMXSIZE*ZSize,img);
		}
	}else{
		for(y=0;y<IMYSIZE;y++)
		for(x=0;x<IMXSIZE;x++)
		for(z=0;z<ZSize;z++) {
#ifdef INVERSE_Y
			P=((IMYSIZE-1-y)*IMXSIZE+x)*ZSize+z;
#else
			P=(y*IMXSIZE+x)*ZSize+z;
#endif
			fscanf(img," %d",&v);
			Data[P]=v;
		}
	}




        fclose(img);
        return(0);
}


/*** Save in PBM Format , If color, Data is RGB RGB RGB ... ***/
int SaveImageDepthMap(char *Name,imginfo *I)
{
FILE *img;
int P,y;
unsigned char *Data;
time_t tm;




	Data=I->Data;

        img=fopen(Name,"w");
        if( img==NULL ) {
                fprintf(stderr,"SaveImage: Unable to open '%s'\n",Name);return(-1);
        }

	 fprintf(img,"P5\n");	/* pgm */


	tm=time(NULL);
	fprintf(img,"# IMG Module, %s",ctime(&tm));

	fprintf(img,"%d %d\n255\n",IMXSIZE,IMYSIZE);


        for(y=0;y<IMYSIZE;y++) {
#ifdef INVERSE_Y
                P=(IMYSIZE-1-y)*IMXSIZE; /* A L'ENVERS */
#else
                P=y*IMXSIZE;
#endif
		fwrite(Data+P,1,IMXSIZE,img);
        }
        fclose(img);

        return(0);
}


