/*autor:  Marc-Antoine Drouin
  email: drouim@iro.umontreal.ca

  description: Compute depth map ultra fast for multiple-baseline
               configuration using IDP with hybrid (geo-coherent VS photo-cohenrent)
               visibility approach.
  
  Copyright (c) 2005 Marc-Antoine Drouin


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



*/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "const.h"		/* parameter of the methode */
#include "img.h"		/* image module use to read and write in ppm and pgm format */
#include "dynamic.h" 		/* dynamic programming module */


/* what we call one iteration is 2 in the code here */
short nbIteration=ITERATION*2 ;
/* smoothing parameter */
OPT_TYPE  smooth= LAMBDA; 	/* disparity smoothing */
OPT_TYPE  gama= GAMA; 		/* visibility mask smoothing */


/* scan from left to right or from right to left 
samething for top bottom and bottom top 
set to one to change order 
The algorithme does change does values*/
unsigned char  reverseOrderX=0;
unsigned char  reverseOrderY=0;
imginfo f; 			/* depth map */



/* visibility information */
#ifdef KEEP_VISIBILITY
imginfo visi;
#endif


#ifdef VISIBILITY_2D
/* this is used to keep track of visibility outside of epipolar line */
OPT_ACC_TYPE visi2d[IMYSIZE][IMXSIZE];
imginfo visi2;
#endif

imginfo il, 			/* left image */
  i,				/* center image (reference image) */
  ir,				/* right image */
  it,				/* top image */
  ib; 				/* bottom image */

/* matching volume DSI  section 3.2 of working document*/
/* indexing order is different because access locality is different between left right DSI and bottom and top DSI */
OPT_TYPE d1[IMYSIZE][IMXSIZE][ NBLABEL]; /* DSI left */
OPT_TYPE d2[IMYSIZE][IMXSIZE][ NBLABEL]; /* DSI right */
OPT_TYPE d3[IMXSIZE][IMYSIZE][ NBLABEL]; /* DSI bottom */
OPT_TYPE d4[IMXSIZE][IMYSIZE][ NBLABEL]; /* DSI top */







/* this is used for nakamura */
#ifdef USED_NAKAMURA
OPT_TYPE dNAKA[IMYSIZE][IMXSIZE][ NBLABEL];
#endif





/* interpolate image used to compute DSI value*/
/* access is as folloing   acceding pixel (x+.5,y)  di[x][y][0] */
/*                                        (x-.5,y)  di[x-1][y][0] */
/*                                        (x,y+.5)  di[x][y+1][1]   */
/*                                        (x,y-.5)  di[x][y-1][1]   */
OPT_TYPE di[IMYSIZE][IMXSIZE][IMZSIZE][2]; /* reference  image  (or center) */
OPT_TYPE dil[IMYSIZE][IMXSIZE][IMZSIZE][2]; /* left image */
OPT_TYPE dir[IMYSIZE][IMXSIZE][IMZSIZE][2]; /* rigth image */
OPT_TYPE dit[IMYSIZE][IMXSIZE][IMZSIZE][2]; /* top image */
OPT_TYPE dib[IMYSIZE][IMXSIZE][IMZSIZE][2]; /* bottom image */

/* smoothing g function  section 3.1 of working document */
/* access is as fallow:  g(x,y  , x+1,y) is in  sX[x][y]   */
/*                       g(x,y  , x-1,y) is in  sX[x-1][y]  */
/*                       g(x-1,y  , x,y) is in  sX[x-1][y]  */
/*                       g(x+1,y  , x,y) is in  sX[x+1][y]  */

/* access is as fallow:  g(x,y  , x,y+1) is in  sY[x][y]   */
/*                       g(x,y  , x,y-1) is in  sY[x][y-1]  */
/*                       g(x,y-1  , x,y) is in  sY[x][y-1]  */
/*                       g(x,y+1  , x,y) is in  sY[x][y+1]  */
OPT_TYPE sX[IMXSIZE][IMYSIZE];
OPT_TYPE sY[IMXSIZE][IMYSIZE];

/* first iteration we do not use smoothing from other line */
unsigned char first=1;

/* used to keep curent colomne or line when optimizing line or colonm */
short Cy; 			/* place Y when optimisation in X */
short Cx;			/* place X when optimisation in Y  */


void licence()
{
printf("HybridIDP version 1.1, Copyright (C) 2005  Marc-Antoine Drouin  and Sebastien Roy\n");
printf("Stereo module comes with ABSOLUTELY NO WARRANTY; for details look in LICENCE.TXT\n");
printf("This is free software, and you are welcome to redistribute it\n");
printf("under certain conditions; for details  look in LICENCE.TXT.\n");
printf("For more details:\n");
printf("Marc-Antoine Drouin, Martin Trudeau and Sbastien Roy, Fast Multiple-baseline Stereo with Occlusion, 3DIM, Ottawa, Canada, June 2005.\n");
}


/* this is the matching cost function  
based on Birchfield and tomasi PAMI paper and Kologorov modification also PAMI dewscribe in section 
3.2 of working document and known as function CostBirchfield in document*/
OPT_TYPE cmpPixel(imginfo * i1,	/* image one */
		  imginfo *i2, 	/* second image */
		  OPT_TYPE  di1[IMYSIZE][IMXSIZE][3][2], /* interpolation information of first image */
		  OPT_TYPE  di2[IMYSIZE][IMXSIZE][3][2], /* interpolation information of second image  */
		  short x1, 	/* coordonate in first image */
		  short y1, 
		  short x2,	/* coordonate in second image */
		  short y2)
{
  OPT_TYPE res; 		/* resulting cost function */
  int z;			/* iterate the canal of image */
  res=0;
  for(z=0;z<IMZSIZE;z++)
    {
      OPT_TYPE c1a,c1b,c1c,c1d,c1e; /* interpolate value  image one */
      OPT_TYPE c2a,c2b,c2c,c2d,c2e; /* interpolate value  image 2 */
      OPT_TYPE tmp1;		/* temporaly variable of finding minimum values */
      OPT_TYPE tmp2;
      OPT_TYPE tmp3;
      OPT_TYPE tmp4;
      OPT_TYPE tmp5;
      OPT_TYPE tmp;

      c1a =2*(OPT_TYPE) GETIMG(*i1,x1,y1,z); /* compensate for interpolation */
      c2a =2*(OPT_TYPE) GETIMG(*i2,x2,y2,z);
    
      /* interpolation value of image 1 */
      c1b= di1[y1][x1][z][0];
      c1d= di1[y1][x1][z][1];
      c1c= di1[y1][x1-1][z][0];
      c1e= di1[y1-1][x1][z][1];
      

      /* interpolation of image 2 */
      c2b= di2[y2][x2][z][0];      
      c2d= di2[y2][x2][z][1];
      c2c= di2[y2][x2-1][z][0];
      c2e= di2[y2-1][x2][z][1];


      tmp  = abs(c1a-c2a );
      tmp1 = abs(c1a-c2b );
      tmp2 = abs(c1a-c2c );
      tmp3 = abs(c1a-c2d );
      tmp4 = abs(c1a-c2e );

      if(tmp2 < tmp)
	tmp = tmp2;
      if(tmp1 < tmp)
	tmp = tmp1;
      if(tmp3 < tmp)
	tmp = tmp3;
      if(tmp4 < tmp)
	tmp = tmp4;
     
       tmp2 = abs(c2a-c1b );
      tmp3 = abs(c2a-c1c );
      tmp4 = abs(c2a-c1d );
      tmp5 = abs(c2a-c1e );
      
      if(tmp3 < tmp)
      	tmp = tmp3;
      if(tmp2 < tmp)
      	tmp = tmp2;
      if(tmp4 < tmp)
      	tmp = tmp4;
      if(tmp5 < tmp)
      	tmp = tmp5;

      if(tmp > 2000  )
	tmp = 2000 ; 		/* maximum values */
      res+=tmp; 
    }
  return res ;
}


/* we fill the interpolation information to be used latter in the cost function  cmpPixel function */
/* we compute the value of intensity of half pixel (i.e. (x+.5,y) (x,y+.5))
  we do not stock the halft value but 2x it to save one division 
*/
void computeInterpolateImage()
{
  short x,y,z;
  unsigned int P;
  unsigned int PZ;
  unsigned int PZ2;
  P=0;
  for(y=0;y<IMYSIZE-1;y++)
    {
      for(x=0;x<IMXSIZE-1;x++)
	  for(z=0;z<IMZSIZE;z++) /* different chanel */
	    {
	      /* interpolation is done directly in loop so we save poiter computation */
	      /* we do not return the interpolation value but 2* interpolate value */
	      PZ=P+IMZSIZE;
	      PZ2= P+IMXSIZE*IMZSIZE;
	      di[y][x][z][0]= i.Data[PZ]+i.Data[P] ;
	      di[y][x][z][1]= i.Data[PZ2]+i.Data[P];
	      dil[y][x][z][0]= il.Data[PZ]+il.Data[P] ;
	      dil[y][x][z][1]= il.Data[PZ2]+il.Data[P];
	      dir[y][x][z][0]= ir.Data[PZ]+ir.Data[P] ;
	      dir[y][x][z][1]= ir.Data[PZ2]+ir.Data[P];
	      dib[y][x][z][0]= ib.Data[PZ]+ib.Data[P] ;
	      dib[y][x][z][1]= ib.Data[PZ2]+ib.Data[P];
	      dit[y][x][z][0]= it.Data[PZ]+it.Data[P] ;
	      dit[y][x][z][1]= it.Data[PZ2]+it.Data[P];
	      P++;
	    }
      P+=IMZSIZE;
    }
}



#ifdef USED_NAKAMURA
/* this is the stuff that we used  */
int cmpNaka (const void * e1, const void * e2 )
{
  OPT_TYPE * a1 = (OPT_TYPE *)e1;
  OPT_TYPE * a2 = (OPT_TYPE * )e2;

  if(*a1==*a2)
    return 0;
  if (*a1 > *a2)
    return 1;
  return -1;

}
#endif






/* Filling the  DSI values  */
/* LABEL_FACTOR is a constant should be one in all case exepte want disparity is compute at step different then one */
/* when no value can be compute because reprojection fall outside of
   supporting image we use maximum possible cost function value to
   flag this situation  */

void computeCostFunction( )
{
  short x,y,z;
  for(y=BORDERY;y<IMYSIZE-BORDERY;y++) /* these a small border for with we do not compute depth  */
    for(x=BORDERX;x<IMXSIZE-BORDERX;x++)
      for(z=0;z<NBLABEL;z++)
	{
	  /* left DSI */
	  if(x+LABEL_FACTOR*z < IMXSIZE)
	    d1[y][x][z]=cmpPixel(&i, &il,di,dil, x,y , x+ LABEL_FACTOR*z,y        );
	  else
	    d1[y][x][z]= OPT_MAX; /* indicate outside of supporting image */
	  /* right DSI */
	  if(x-LABEL_FACTOR*z >= 0)
	    d2[y][x][z]=cmpPixel(&i, &ir,di,dir,  x,  y,  x- LABEL_FACTOR*z,  y   );
	  else
	    d2[y][x][z]= OPT_MAX;	/* indicate outside of supporting image */
	}
  for(x=BORDERX;x<IMXSIZE-BORDERX;x++)
    for(y=BORDERY;y<IMYSIZE-BORDERY;y++) /* these a small border for with we do not compute depth  */
      for(z=0;z<NBLABEL;z++)
	{
		  /* bottom DSI */
	  if(y+LABEL_FACTOR*z < IMYSIZE)
	    d3[x][y][z]=cmpPixel(&i, &ib,di,dib, x,y , x,y +LABEL_FACTOR*z       );
	  else
	    d3[x][y][z]= OPT_MAX;/* indicate outside of supporting image */
	  /* top DSI */
	  if(y-LABEL_FACTOR*z >= 0)
	    d4[x][y][z]=cmpPixel(&i, &it,di,dit,  x,  y,  x,  y- LABEL_FACTOR*z   );
	  else
	    d4[x][y][z]= OPT_MAX;/* indicate outside of supporting image */
	}


  /*  */
#ifdef USED_NAKAMURA

for(y=BORDERY;y<IMYSIZE-BORDERY;y++) /* these a small border for with we do not compute depth  */
    for(x=BORDERX;x<IMXSIZE-BORDERX;x++)
      for(z=0;z<NBLABEL;z++)
	{
	  OPT_TYPE a[4] ; 
	  a[0] =d1[y][x][z];
	  a[1] =d2[y][x][z];
	  a[2] =d3[x][y][z];
	  a[3] =d4[x][y][z];
	  qsort(a,4,sizeof(OPT_TYPE),cmpNaka );
	  
#ifdef NAKAMURA_2D	  
	  dNAKA[y][x][z]=  (a[0]+a[1])/2;
#else
	  dNAKA[y][x][z]= a[0]  ;
#endif
	}
#endif
}



/* compute the smoothing function g */
void computeSmoothFunction()
{
  short x,y;  
  for(y=BORDERY;y<IMYSIZE-BORDERY-1;y++) /* we alway remove part of the reference image for with we do not compute depth*/
    for(x=BORDERX;x<IMXSIZE-BORDERX-1;x++)
      {
	OPT_TYPE diff;
	OPT_TYPE diff2;
	sX[x][y]=smooth; 	/* this is lambda */
	sY[x][y]=smooth;
	/* smoothing is proportinal to image gradiant this come from kolmogorov 2002 paper */
	diff = abs(GETIMG(i,x,y,0)-GETIMG(i,x+1,y,0) )+
	   abs(GETIMG(i,x,y,1)-GETIMG(i,x+1,y,1) )+
	  abs(GETIMG(i,x,y,2)-GETIMG(i,x+1,y,2) );
	
	if(diff <3*5)
	  sX[x][y]+=2* smooth; 		/* we put 3 times lambda  in the x axis*/
	
	diff2 = abs(GETIMG(i,x,y,0)-GETIMG(i,x,y+1,0) )+
	  abs(GETIMG(i,x,y,1)-GETIMG(i,x,y+1,1) )+
	  abs(GETIMG(i,x,y,2)-GETIMG(i,x,y+1,2) );

	if(diff2 <3*5)
	  sY[x][y]+=2* smooth; /* we put 3 times lambda  in the y axis*/
      }
}


#ifdef USED_NAKAMURA
/* this code is not realted to our algorithm only used to compare with Nakamura 96 paper using iterative dynamic programing 04 paper */
/* result are inferior to us */

OPT_TYPE SmoothCapacityNaka( short y1, unsigned char l1, short y2,unsigned char d, int occ, void * v)
{
  short x1,x2;   
  OPT_TYPE scost;		/* smoothing cost oder direction */
  short olx2,oly2,olx1,oly1;
  if(Cy<0)
    { 
      if(reverseOrderY==1)
	{
	  y1= IMYSIZE-BORDERY-y1-1;
	  y2= IMYSIZE-BORDERY-y2-1;
	}
      else
	{
	  y1+=BORDERY;
	  y2+=BORDERY;
	}
      x1=Cx;
      x2=Cx;

      olx1 = x1;
      olx2 = x2;
      oly1 = y1;
      oly2 = y2;

#ifdef ROBUST
      scost = 0;
   if(GETDEPTH(f,Cx-1,y2)!=d)
     scost = sX[Cx-1][y2];
   if(GETDEPTH(f,Cx+1,y2)!=d)
     scost +=  sX[Cx][y2];
#else
   scost = abs(GETDEPTH(f,Cx-1,y2)-d)*sX[Cx-1][y2] +
     abs(GETDEPTH(f,Cx+1,y2)-d)*sX[Cx][y2] ;       
#endif
   if(first==1)
     scost=0;
   if(y1 > y2)
    y1=y2;
  
#ifdef ROBUST
  if(l1!=d)
     scost +=  sY[Cx][y1];
#else
   if(l1>d)
     scost+= (l1-d)*sY[Cx][y1];
   else
     scost+= (d-l1)*sY[Cx][y1];
#endif
    }
  else 
    {
      if(reverseOrderX==1)
	{
	  x1= IMXSIZE-BORDERX-y1-1;
	  x2= IMXSIZE-BORDERX-y2-1;
	}
      else
	{
	  x1=y1+BORDERX;
	  x2=y2+BORDERX;
	}
      y1=Cy;
      y2=Cy; 
      
      olx1 = x1;
      olx2 = x2;
      oly1 = y1;
      oly2 = y2;
   
#ifdef ROBUST
      scost=0;
      if(GETDEPTH(f,x2,Cy-1)!=d)
	scost = sY[x2][Cy-1];
      if(GETDEPTH(f,x2,Cy+1)!=d)
	scost += sY[x2][Cy];
#else
      scost = abs(GETDEPTH(f,x2,Cy-1)-d)*sY[x2][Cy-1]  + 
	abs(GETDEPTH(f,x2,Cy+1)-d)*sY[x2][Cy];
#endif
   if(first==1)
     scost=0;
   if(x1 > x2)
     x1=x2;

#ifdef ROBUST
   if(l1!=d)
     scost += sX[x1][Cy];
#else
   if(l1>d)
     scost+= (l1-d)*sX[x1][Cy];
   else
     scost+= (d-l1)*sX[x1][Cy];
#endif
    }

    return (scost+ dNAKA[oly2][olx2][d])*8;
}
#endif



/* this is the function used for likelihood when working on colomn */
/* this fonction compute the likellehood term base of visibility */
/* this function take about 30% of computing power  */
#ifndef KEEP_VISIBILITY
OPT_TYPE SmoothCapacityY( short y1, unsigned char l1, short y2,unsigned char d, int occ, void * v)
#else
OPT_TYPE SmoothCapacityY( short y1, unsigned char l1, short y2,unsigned char d, int occ,MASK_TYPE oldmask,MASK_TYPE * mask ,void * v)
#endif
{
  OPT_TYPE res1= OPT_MAX; 	/* this is a flag to indicate that
				   cost function not avalable because
				   of occlusion or comparaison
				   impossible du to pixel outside of
				   image */  
  OPT_TYPE res2  =  OPT_MAX;
  OPT_TYPE scost;		/* smoothing cost horizontal (idp) direction */  


  /* this is the initialisation */
  if(y1<0)
    {
      if(reverseOrderY==1)
	return  d1[IMYSIZE-BORDERY-y2-1][Cx][d] ; 		/* left */
      else
	return  d2[y2+BORDERY][Cx][d] ; 		/* rigth */
    }


  if(reverseOrderY==1)  	/* are we working top toward bottom or the orther way */
    {
      y1= IMYSIZE-BORDERY-y1-1;
      y2= IMYSIZE-BORDERY-y2-1;
    }
  else
    {
      y1+=BORDERY;
      y2+=BORDERY;
    }






  /* we change camera */
  if(reverseOrderY )
    {
      if (occ<d  ) /* visibility test to select mask */
	{	  
#ifdef VISIBILITY_2D
	  /* we also check for the top DSI (4) */      
	  if(visi2d[y2][Cx-1] < d+Cx && first!=1 )  
	    res1= (d1[y2][Cx][d]+d4[Cx][y2][d])/2  ;	/* left and botom DSI */
	  else
	    res1=  d4[Cx][y2][d]  ; /* bottom DSI */
#else
      	  res1=  d4[Cx][y2][d]  ; /* bottom DSI */
#endif
	}
      /* camera along the axis  we may have to gest*/
#ifdef VISIBILITY_2D 
      if(occ >= d)
	{
	  if(  visi2d[y2][Cx-1] < d+Cx && first!=1   )
	    res2 = d1[y2][Cx][d]  ; 	/* top DSI */

	  else if( d2[y2][Cx][d] < d3[Cx][y2][d])
	    res2 =  d2[y2][Cx][d]; /* nakamura */
	  else
	    res2=d3[Cx][y2][d]  ; 	/* nakamura */  

	}
#else   
      if (occ>=d)
	res2=d3[Cx][y2][d]  ; 	/* top DSI */
#endif
    }
  else 				/* no reverse order */
    { 
      if( occ<d) /* since opt order is change visibility check is inverted */
	{  
#ifdef VISIBILITY_2D
	  /* we also check for the top DSI (4) */      
	  if(visi2d[y2][Cx-1] < d+Cx && first!=1 )
	    res1= (d1[y2][Cx][d]+d3[Cx][y2][d])/2  ;	/* left and botom DSI */
	  else
	    res1= d3[Cx][y2][d];    /* bottom DSI */
#else     
	  res1= d3[Cx][y2][d];    /* bottom DSI */
#endif
	}
      /* other camera */
#ifdef VISIBILITY_2D 
      if(occ >= d)
	{
	  if(  visi2d[y2][Cx-1] < d+Cx && first!=1   )
	    res2 = d1[y2][Cx][d]  ; 	/* top DSI */

	  else if( d2[y2][Cx][d] < d4[Cx][y2][d])
	    res2 =  d2[y2][Cx][d];
	  else
	    res2=d4[Cx][y2][d]  ; 	/* top DSI */

	}
      
#else   
      if (occ>=d  )
	res2=d4[Cx][y2][d]  ; 	/* top DSI */
#endif
    }
  
  if(res1== OPT_MAX && res2== OPT_MAX) /* impossible to select mask other than (0,0) */
    return 8000 ;
#ifdef ROBUST 			/* pot smoothing using horizontal value from previous computation */
  scost = 0;
  if(GETDEPTH(f,Cx-1,y2)!=d)
    scost = sX[Cx-1][y2];
  if(GETDEPTH(f,Cx+1,y2)!=d)
    scost +=  sX[Cx][y2];
#else  /* linear model */
  scost = abs(GETDEPTH(f,Cx-1,y2)-d)*sX[Cx-1][y2] +
    abs(GETDEPTH(f,Cx+1,y2)-d)*sX[Cx][y2] ;       
#endif
  if(first==1) 		/* first computation we use ordinary dp since no previous partial solution avalable */
    scost=0;
  if(y1 > y2)
    y1=y2;
    /* smoothing in this line for pot  */
#ifdef ROBUST
  if(l1!=d)
    scost +=  sY[Cx][y1];
#else
  if(l1>d)
    scost+= (l1-d)*sY[Cx][y1];
  else
    scost+= (d-l1)*sY[Cx][y1];
#endif
  /* we select the proper visibility mask and return value to optimisation module */
  if(res1== OPT_MAX)
    {
#ifdef KEEP_VISIBILITY
      *mask = 0;
      if(*mask != oldmask)
	res2+= gama;
#endif
      return (res2+scost)*8 ;
    }
#ifdef KEEP_VISIBILITY
  *mask = 1;
  if(*mask != oldmask)
    res1+= gama;
#endif
  return (res1+scost)*8;
}


#ifndef KEEP_VISIBILITY
OPT_TYPE SmoothCapacityX( short x1, unsigned char l1, short x2,unsigned char d, int occ, void * v)
#else
OPT_TYPE SmoothCapacityX( short x1, unsigned char l1, short x2,unsigned char d, int occ,MASK_TYPE oldmask, MASK_TYPE * mask ,void * v)
#endif
{
    OPT_TYPE  res1= OPT_MAX; 	/* left DSI */
    OPT_TYPE  res2= OPT_MAX;	/* right DSI */
   
    
    OPT_TYPE  scost;		/* smoothing cost oder direction */   



  /* this is the initialisation */
    if(x1<0)
      {
	if(reverseOrderX==1)
	  return  d1[IMXSIZE-BORDERX-x2-1][Cy][d] ; 		/* left */
	else
	  return  d2[x2+BORDERX][Cy][d] ; 		/* rigth */
      }


       if(reverseOrderX==1)
	 {
	   x1= IMXSIZE-BORDERX-x1-1;
	   x2= IMXSIZE-BORDERX-x2-1;
	 }
       else
	 {
	   x1+=BORDERX;
	   x2+=BORDERX;
	 }
      
 /* we chnage camera */
      if(reverseOrderX )
	{
	  if (occ<d ) 		/* >= */
	    {
#ifdef VISIBILITY_2D
	      /* we also check for the top DSI (4) */
	      if(visi2d[Cy-1][x2] < d+Cy && first!=1  )
		res1= (d2[Cy][x2][d]+d3[x2][Cy][d])/2  ;	/* left and top DSI */
	      else
		res1= d2[Cy][x2][d];
#else
	      res1= d2[Cy][x2][d];
#endif
	    }
#ifdef VISIBILITY_2D
	  if(occ >= d)
	    {
	      if( visi2d[Cy-1][x2] < d+Cy && first!=1 )
		res2 = d3[x2][Cy][d];  


	      else if(d4[x2][Cy][d] < d1[Cy][x2][d])
		res2= d4[x2][Cy][d];
	      else
		res2=d1[Cy][x2][d];	/* right DSI */

	    }
#else
	  if( occ>=d )
	    res2=d1[Cy][x2][d];	/* right DSI */
#endif
	}
      else 			/* not reverse order */
	{
	  if( occ<d  )
	    {
#ifdef VISIBILITY_2D
	      if(visi2d[Cy-1][x2] < d+Cy && first!=1  )  
		res1= (d1[Cy][x2][d]+d3[x2][Cy][d])/2  ;	/* left and top DSI */
	      else
		res1=d1[Cy][x2][d]; /* left DSI */
#else
	      res1=d1[Cy][x2][d]; /* left DSI */
#endif
	    }
	  
#ifdef VISIBILITY_2D
	  
	  if(occ >=d)
	    {
	      
	      if( visi2d[Cy-1][x2] < d+Cy && first!=1 )
		res2 = d3[x2][Cy][d];


	      else if(d4[x2][Cy][d] < d2[Cy][x2][d])
		res2= d4[x2][Cy][d];
	      else
		res2=d2[Cy][x2][d];	/* right DSI */

	    }
#else
	  if( occ>=d )
	    res2=d2[Cy][x2][d];	/* right DSI */
#endif  
	}
      

      if(res1== OPT_MAX && res2== OPT_MAX)
	return 8000 ;
      
#ifdef ROBUST
      scost=0;
      if(GETDEPTH(f,x2,Cy-1)!=d)
	scost = sY[x2][Cy-1];
      if(GETDEPTH(f,x2,Cy+1)!=d)
	scost += sY[x2][Cy];
#else
      scost = abs(GETDEPTH(f,x2,Cy-1)-d)*sY[x2][Cy-1]  + 
	abs(GETDEPTH(f,x2,Cy+1)-d)*sY[x2][Cy];
#endif


      if(first==1)
	scost=0;



      if(x1>x2)
	x1=x2;
      




#ifdef ROBUST
      if(l1!=d)
	scost += sX[x1][Cy];
   
#else
      if(l1>d)
	scost+= (l1-d)*sX[x1][Cy];
      else
	scost+= (d-l1)*sX[x1][Cy];
#endif
 
      if(res1== OPT_MAX)
	{
	  
#ifdef KEEP_VISIBILITY
	  *mask = 0;
	  if(*mask != oldmask)
	    res2+= gama;

	  
#endif  

	  return (res2+scost)*8;
	}

      
#ifdef KEEP_VISIBILITY
      *mask = 1;
      if(*mask != oldmask)
	res1+= gama;

#endif   

      return (res1+scost)*8;     
}








/* this is the main loop that implement the iterative dynamic programic using my generic dp module */
int optimize5images()
{
  int result;
  short y,x,j;
  struct dynamicProg  dy,dx;

#if IMXSIZE > IMYSIZE1
  unsigned char res[IMXSIZE];
#else
  unsigned char res[IMYSIZE];
#endif
  int change;
  unsigned char tmp;
  result=0;
  /* init 2 dp module */
  createDynamicProg(&dy,
		    IMXSIZE-2*BORDERX /* length of the line */
		    );
  
  createDynamicProg(&dx,
		    IMYSIZE-2*BORDERY /* length of the line */		   
		    ); 
  
  first=1; 			/* for the first depth map computation do not used idp smoothing */
  for(j=0;j<nbIteration;j++)
    {
      /* we reverse opt orther so we have l->r  t->b  r->l and b ->t */
      if(reverseOrderY==1)
	reverseOrderY=0;
      else
	reverseOrderY=1;
      if(reverseOrderX==1)
	reverseOrderX=0;
      else
	reverseOrderX=1;
      Cx=-1;
      change=0;
#ifdef VISIBILITY_2D
      /* we init the 2D visibility */
      for(x=BORDERX;x<IMXSIZE-BORDERX;x++)
	  visi2d[BORDERY-1][x]=BORDERY-1;
#endif
      for(y=BORDERY;y<IMYSIZE-BORDERY;y++)
	{
	  Cy=y; 		/* curent colomn use in cost function */
	  /* do the optimisation of the problem define in <d> */
	  OptimizeDynamicProg(&dy, 0,NULL );
	  /*  extract the argument of the resulting minimum solution*/
	  ResultDynamicProg(&dy,res   , IMXSIZE-2*BORDERX );
	  /* we copy back in current depth map the solution for this line */
	  if(reverseOrderX)
	    {
	      for(x=0;x<IMXSIZE-2*BORDERX;x++)
		{
		  tmp =  res[IMXSIZE-2*BORDERX-x-1];
		  if(tmp != 	GETDEPTH(f,x+BORDERX,Cy))
		    {
		      change++;
		      GETDEPTH(f,x+BORDERX,Cy)=tmp;
		    }
		}
	    }
	  else
	    {
	      for(x=0;x<IMXSIZE-2* BORDERX;x++)
		{
		  if(res[x]!= 	GETDEPTH(f,x+BORDERX,Cy))
		    {
		      change++;
		      GETDEPTH(f,x+BORDERX,Cy)  =   res[x];
		    }
		}
	    }


#ifdef VISIBILITY_2D
	  for(x=BORDERX;x<IMXSIZE-BORDERX;x++)
	    if(visi2d[Cy-1][x]<= GETDEPTH(f,x,Cy)+Cy  )
	      visi2d[Cy][x]= GETDEPTH(f,x,Cy)+Cy;
	    else
	      visi2d[Cy][x]=visi2d[Cy-1][x];
#endif


#ifdef KEEP_VISIBILITY
	  {
#if IMXSIZE > IMYSIZE1
	    MASK_TYPE res2[IMXSIZE];
#else
	    MASK_TYPE res2[IMYSIZE];
#endif
	    ResultVisibilityDynamicProg(&dy,res2   , IMXSIZE-2*BORDERX );
	  /* we copy back in current depth map the solution for this line */
	  if(reverseOrderX)
	    {
	      for(x=0;x<IMXSIZE-2*BORDERX;x++)
		{
		  GETDEPTH(visi,x+BORDERX,Cy)= res2[IMXSIZE-2*BORDERX-x-1];
		}
	    }
	  else
	    for(x=0;x<IMXSIZE-2* BORDERX;x++)
	      {
		GETDEPTH(visi,x+BORDERX,Cy)  =   res2[x];
	      }
	  }
#endif
	}


      



      result++;
      if(change==0) 		/* if convergence acheived return should never append since prouf of non-convergence exist */
	break;
      first=0; 			/* at this point we are sur that a previous depth map exist so full smoothing of idp */
      Cy=-1; 
      change =0;


#ifdef VISIBILITY_2D
   /* we init the 2D visibility */
      for(y=BORDERY;y<IMYSIZE-BORDERY;y++)
	  visi2d[y][BORDERX-1]=BORDERX-1;
#endif

      for(x=BORDERX;x<IMXSIZE-BORDERX;x++) /* for each colom we do opt */
	{
	    Cx=x;		/* curent line */
	  /* do the optimisation of the problem define in <d> */
	  OptimizeDynamicProg(&dx , 1,NULL);


	  /*  extract the argument of the resulting minimum solution*/
	  ResultDynamicProg(&dx,res   , IMYSIZE-2*BORDERY );


	  /* we copy back in current depth map the solution for this line */
	  if(reverseOrderY)
	    for(y=0;y<IMYSIZE-2*BORDERY;y++)
	      {
		tmp  = res[IMYSIZE-2*BORDERY-y-1];
		if(	GETDEPTH(f,Cx,y+BORDERY)  !=  tmp)
		  {
		    GETDEPTH(f,Cx,y+BORDERY)  =  tmp;
		    change++;
		  }
	      }
	  else
	    for(y=0;y<IMYSIZE-2* BORDERY;y++)	
	      {
		if(res[y]!= GETDEPTH(f,Cx,y+BORDERY))
		  {
		    GETDEPTH(f,Cx,y+BORDERY)  =   res[y]; 
		    change++;
		  }
	      }

#ifdef VISIBILITY_2D
		for(y=BORDERY;y<IMYSIZE-BORDERY;y++)
		  if(visi2d[y][Cx-1]  <=   GETDEPTH(f,Cx,y)+Cx  )
		    visi2d[y][Cx]=GETDEPTH(f,Cx,y)+Cx;		
		  else
		    visi2d[y][Cx]=visi2d[y][Cx-1];		   
#endif




#ifdef KEEP_VISIBILITY
	  {
	    
#if IMXSIZE > IMYSIZE1
	    MASK_TYPE res2[IMXSIZE];
#else
	    MASK_TYPE res2[IMYSIZE];
#endif   
	    /*  extract the argument of the resulting minimum solution*/
	    ResultVisibilityDynamicProg(&dx,res2   , IMYSIZE-2*BORDERY );

	    /* we copy back in current depth map the solution for this line */
	    if(reverseOrderY)
	      for(y=0;y<IMYSIZE-2*BORDERY;y++)
		{
		  GETDEPTH(visi,Cx,y+BORDERY)  =  res2[IMYSIZE-2*BORDERY-y-1];
		}
	    else
	      for(y=0;y<IMYSIZE-2* BORDERY;y++)	
		{
		      GETDEPTH(visi,Cx,y+BORDERY)  =   res2[y];
		} 
	  }
#endif
	}


      result++;
      if(change==0)
	break;

      
      
      
    }    


  freeDynamicProg(&dy);
  freeDynamicProg(&dx);
  return result;
}






/* we remove noise from depth map */
/* this is a simple filter */
void postProcessing()
{
  short x,y;
  for(y=BORDERY+1;y<IMYSIZE-BORDERY-1;y++)
    {
    for(x=BORDERX+1;x<IMXSIZE-BORDERX-1;x++)
      {
	if(GETDEPTH(f,x-1,y)== GETDEPTH(f,x+1,y) && GETDEPTH(f,x-1,y)!=GETDEPTH(f,x,y))
	  GETDEPTH(f,x,y)=GETDEPTH(f,x-1,y);
	else if(GETDEPTH(f,x,y-1)== GETDEPTH(f,x,y+1) && GETDEPTH(f,x,y-1)!=GETDEPTH(f,x,y))
	  GETDEPTH(f,x,y)=GETDEPTH(f,x,y-1);
      }
    }
}







/* main access point 
parameter should be define in const.h */


int main()
{  
  short x,y;
  int k;
  int res;
  /* do we have ground truth for erreor computation
     if no we do not even dump intermediar depth map on disk */
#ifdef GROUNDTRUE
  imginfo true;  
#endif

  licence();

  if (LoadImage(IMAGE_LEFT ,&il) ||   LoadImage(IMAGE_REF, &i)  || LoadImage(IMAGE_RIGHT,&ir) || LoadImage(IMAGE_BOTTOM,&ib) || LoadImage(IMAGE_TOP,&it) )
    {
      printf("Unable to read image\n");
      return 1;
    }


#ifdef GROUNDTRUE
  
  if (LoadImage(GROUNDTRUE,&true))
    {
      printf("Unable to read ground true\n");
      return 1;
    }
#endif


 
  printf("Interpolating image\n");
  computeInterpolateImage();
  printf("Computing DSI\n");
  computeCostFunction( );
  printf("optimize\n");
  reverseOrderX=0;
  reverseOrderY=1;
  computeSmoothFunction();          
  res= optimize5images();
  postProcessing();
  

   /* we scale */
  for(k=0;k<IMXSIZE*IMYSIZE;k++)
    f.Data[k]*=FACTOR_SAVE;
  /* error computation and dumping on disk of depth map */
#ifdef GROUNDTRUE
  /* compute error */
  {
    int err;
    int total;
    err=total=0;
    for(y=18;y<IMYSIZE-18;y++)
      for(x=18;x<IMXSIZE-18;x++)
	{
	  if( abs((int)GETDEPTH(f,x,y)  -(int)(GETDEPTH(true,x,y)))  >(int)FACTOR_SAVE)
	    err++;
	  total++;
	}
    printf("Error is %f \n",(float)err/(float)total);
  }    
#endif     
  printf("Depth map saved under res.pgm\n");
  SaveImageDepthMap("res.pgm",&f); /* final solution */
  return 0;
}
